/* systemMenu.js
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

/* exported getExtension */

const Gettext = imports.gettext;

const { Gio, GLib, GObject, Gvc, Meta, Shell, St } = imports.gi;

const ExtensionUtils = imports.misc.extensionUtils;
const LoginManager = imports.misc.loginManager;
const Util = imports.misc.util;

const BoxPointer = imports.ui.boxpointer;
const Main = imports.ui.main;
const PanelMenu = imports.ui.panelMenu;
const PopupMenu = imports.ui.popupMenu;

const StatusSystem = imports.ui.status.system;
const StatusVolume = imports.ui.status.volume;

const Me = ExtensionUtils.getCurrentExtension();

const _ = Gettext.domain(Me.metadata['gettext-domain']).gettext;

const BASIC_SETTINGS = [
    'ca.desrt.dconf-editor.desktop',        // dconf-editor
    'gconf-editor.desktop',                 // gconf-editor
    'gnome-control-center.desktop',         // gnome-control-center
    'org.gnome.Extensions.desktop',         // gnome-shell-extension-prefs
    'org.gnome.tweaks.desktop',             // gnome-tweaks
    'system-config-printer.desktop',        // system-config-printer
    'synaptic.desktop',                     // synaptic
    'users.desktop'                         // gnome-system-tools
]

const GNOME_TERMINAL = 'org.gnome.Terminal.desktop';

const KEY_ALWAYS_SHOW_AUDIO_INPUT = 'always-show-audio-input';
const KEY_DISABLE_RESTART_BUTTONS = 'disable-restart-buttons';
const KEY_SHOW_SYSTEMMENU = 'show-systemmenu';

const SYSTEMMENU_DEFAULT = 0;
const SYSTEMMENU_ICON = 1;

const PopupSettingsMenuItem = GObject.registerClass(
class PopupSettingsMenuItem extends PopupMenu.PopupImageMenuItem {
    _init() {
        super._init(_("Settings Panel Manager"), 'preferences-other-symbolic');
    }

    activate(event) {
        Util.spawnCommandLine('gnome-shell-extension-prefs '+Me.metadata['uuid']);
        Main.overview.hide();
        super.activate(event);
    }
});

const PopupTerminalMenuItem = GObject.registerClass(
class PopupTerminalMenuItem extends PopupMenu.PopupImageMenuItem {
    _init() {
        super._init('', '');
        this._appSystem = Shell.AppSystem.get_default();
        this._appSystemChangedId = this._appSystem.connect('installed-changed', this._onInstalledChanged.bind(this));
        this._onInstalledChanged();
    }

    _onDestroy() {
        this._appSystem.disconnect(this._appSystemChangedId);
        super._onDestroy();
    }

    _onInstalledChanged() {
        this._app = this._appSystem.lookup_app('org.gnome.Terminal.desktop');
        if (this._app) {
            this.label.set_text(this._app.get_name());
            this.setIcon(this._app.get_icon());
            this.show();
        }
        this.visible = this._app;
    }

    activate(event) {
        this._app.activate();
        Main.overview.hide();
        super.activate(event);
    }
});

const PopupHibernationMenuSection = class extends PopupMenu.PopupMenuSection {
    constructor() {
        super();
        this._itemHybridSleep = new PopupMenu.PopupMenuItem(_("Hybrid Sleep"));
        this._itemHybridSleep.connect('activate', () => {
            this._onHibernation('HybridSleep');
        });
        this.addMenuItem(this._itemHybridSleep);
        this._itemHibernate = new PopupMenu.PopupMenuItem(_("Hibernate"));
        this._itemHibernate.connect('activate', () => {
            this._onHibernation('Hibernate');
        });
        this.addMenuItem(this._itemHibernate);
        this._loginManager = LoginManager.getLoginManager();
        this._openStateChangedId = Main.panel.statusArea.aggregateMenu._system.menu.connect('open-state-changed', (menu, open) => {
            if (!open) return;
            this._updateHibernation();
        });
    }

    _canHibernation(value, asyncCallback) {
        if (this._loginManager._proxy) {
            this._loginManager._proxy.call(value, null, Gio.DBusCallFlags.NONE, -1, null, (proxy, asyncResult) => {
                try {
                    let result = proxy.call_finish(asyncResult).deep_unpack();
                    let disableRestartButton = Main.panel.statusArea.aggregateMenu._system._systemActions._loginScreenSettings.get_boolean(KEY_DISABLE_RESTART_BUTTONS);
                    let disabled = Main.sessionMode.isLocked || (Main.sessionMode.isGreeter && disableRestartButton);
                    asyncCallback(result[0] !== 'no' && !disabled);
                } catch(e) {
                    asyncCallback(false);
                }
            });
        } else {
            Mainloop.idle_add(() => {
                asyncCallback(false);
                return false;
            });
        }
    }

    _onHibernation(value) {
        Main.overview.hide();
        if (this._loginManager._proxy) {
            this._loginManager._proxy.call(value, GLib.Variant.new('(b)', [true]), Gio.DBusCallFlags.NONE, -1, null, null);
        } else {
            this._loginManager.emit('prepare-for-sleep', true);
            this._loginManager.emit('prepare-for-sleep', false);
        }
    }

    _updateHibernation() {
        this._canHibernation("CanHibernate", (result) => {
            this._itemHibernate.actor.visible = result;
        });
        this._canHibernation("CanHybridSleep", (result) => {
            this._itemHybridSleep.actor.visible = result;
        });
    }

    destroy() {
        Main.panel.statusArea.aggregateMenu._system.menu.disconnect(this._openStateChangedId);
        super.destroy();
    }
};

const PopupPreferencesSubMenuMenuItem = GObject.registerClass(
class PopupPreferencesSubMenuMenuItem extends PopupMenu.PopupSubMenuMenuItem {
    _init() {
        super._init(_("Basic Settings"), true);
        this.icon.icon_name = 'preferences-system-symbolic';
        this._appSystem = Shell.AppSystem.get_default();
        this._appSystemChangedId = this._appSystem.connect('installed-changed', this._reloadPreferences.bind(this));
        this._reloadPreferences();
    }

    _reloadPreferences() {
        this.menu.removeAll();
        let appList = [];
        BASIC_SETTINGS.forEach(desktopFile => {
            let app = this._appSystem.lookup_app(desktopFile);
            if (app) appList.push(app);
        }, this);
        appList.sort((a, b) => {
            return a.compare_by_name(b);
        });
        appList.forEach(app => {
            let name = app.get_name();
            let item = new PopupMenu.PopupMenuItem(name);
            item.connect('activate', () => {
                app.activate();
                Main.overview.hide();
            });
            this.menu.addMenuItem(item);
        }, this);
        this.actor.visible = this.menu.numMenuItems > 0;
    }

    destroy() {
        this._appSystem.disconnect(this._appSystemChangedId);
        super.destroy();
    }
});

const PopupVolumeSubMenuMenuItem = GObject.registerClass(
class PopupVolumeSubMenuMenuItem extends PopupMenu.PopupSubMenuMenuItem {
    _init(deviceType) {
        super._init('...', true);
        this._controlChangedId = [
            Main.panel.statusArea.aggregateMenu._volume._control.connect('state-changed', this._updateDevices.bind(this)),
            Main.panel.statusArea.aggregateMenu._volume._control.connect('%s-added'.format(deviceType), this._updateDevices.bind(this)),
            Main.panel.statusArea.aggregateMenu._volume._control.connect('%s-removed'.format(deviceType), this._updateDevices.bind(this)),
            Main.panel.statusArea.aggregateMenu._volume._control.connect('active-%s-update'.format(deviceType), this._setActiveDevice.bind(this)),
        ];
        this._updateDevices();
    }

    _addDevice(id) {
        let uidevice = this.getDeviceById(id);
        if (!uidevice || !uidevice.port_available) return null;
        let deviceInfo = new Object();
        deviceInfo.uidevice = uidevice;
        deviceInfo.text = uidevice.description;
        if (uidevice.origin !== '') deviceInfo.text += ' - '+uidevice.origin;
        deviceInfo.item = this.menu.addAction(deviceInfo.text, () => {
            this.changeDevice(uidevice);
        });
        this._devices[id] = deviceInfo;
        return uidevice;
    }

    _setActiveDevice(control, id) {
        let deviceInfo = this._devices[id];
        if (deviceInfo && deviceInfo !== this._activeDevice) {
            if (this._activeDevice) this._activeDevice.item.setOrnament(PopupMenu.Ornament.NONE);
            this._activeDevice = deviceInfo;
            deviceInfo.item.setOrnament(PopupMenu.Ornament.CHECK);
            this.label.text = deviceInfo.text;
            let icon = deviceInfo.uidevice.get_icon_name();
            if (icon === null || icon.trim() === '') icon = this.getDefaultIcon();
            this.icon.icon_name = icon;
        }
    }

    _updateDevices() {
        this.menu.removeAll();
        this._devices = {};
        if (Main.panel.statusArea.aggregateMenu._volume._control.get_state() === Gvc.MixerControlState.READY) {
            let defaultStream = this.getDefaultStream();
            let numId = new Gvc.MixerUIDevice().get_id();
            for (let id = 0; id <= numId; id++) {
                let uidevice = this._addDevice(id);
                if (uidevice) {
                    let stream = Main.panel.statusArea.aggregateMenu._volume._control.get_stream_from_device(uidevice);
                    if (stream) {
                        let stream_port = stream.get_port();
                        if (stream_port && stream === defaultStream && stream_port.port === uidevice.get_port()) this._setActiveDevice(Main.panel.statusArea.aggregateMenu._volume._control, id);
                    }
                }
            }
        }
        this._updateVisibility();
    }

    _updateVisibility() {
        this.actor.visible = this.menu.numMenuItems > 1;
    }

    destroy() {
        this._controlChangedId.forEach(id => {
            Main.panel.statusArea.aggregateMenu._volume._control.disconnect(id);
        });
        super.destroy();
    }
});

const PopupVolumeInputSubMenuMenuItem = GObject.registerClass(
class PopupVolumeInputSubMenuMenuItem extends PopupVolumeSubMenuMenuItem {
    _init() {
        super._init('input');
        Main.panel.statusArea.aggregateMenu._volume._volumeMenu._input._maybeShowInput = function() {
            let alwaysShow = ExtensionUtils.getSettings().get_boolean(KEY_ALWAYS_SHOW_AUDIO_INPUT);
            if (alwaysShow) {
                this._showInput = true;
                this._updateVisibility();
            } else {
                StatusVolume.InputStreamSlider.prototype._maybeShowInput.call(this);
            }
        };
        Main.panel.statusArea.aggregateMenu._volume._volumeMenu._input._shouldBeVisible = function() {
            let alwaysShow = ExtensionUtils.getSettings().get_boolean(KEY_ALWAYS_SHOW_AUDIO_INPUT);
            if (alwaysShow) {
                return this._stream !== null;
            } else {
                StatusVolume.InputStreamSlider.prototype._shouldBeVisible;
            }
        };
        this._settings = ExtensionUtils.getSettings();
        this._settingsChangedId = this._settings.connect('changed::%s'.format(KEY_ALWAYS_SHOW_AUDIO_INPUT), this._updateVisibility.bind(this));
        if (Main.layoutManager._startingUp) {
            this._startupComplete = Main.layoutManager.connect('startup-complete', () => {
                this._updateVisibility();
                Main.layoutManager.disconnect(this._startupComplete);
            });
        } else {
            this._updateVisibility();
        }
    }

    _updateVisibility() {
        Main.panel.statusArea.aggregateMenu._volume._volumeMenu._input._maybeShowInput.call(Main.panel.statusArea.aggregateMenu._volume._volumeMenu._input);
        super._updateVisibility();
        let showInput = Main.panel.statusArea.aggregateMenu._volume._volumeMenu._input._showInput;
        if (!showInput) this.actor.hide();
    }

    changeDevice(uidevice) {
        Main.panel.statusArea.aggregateMenu._volume._control.change_input(uidevice);
    }

    destroy() {
        Main.panel.statusArea.aggregateMenu._volume._volumeMenu._input._shouldBeVisible = StatusVolume.InputStreamSlider.prototype._shouldBeVisible;
        Main.panel.statusArea.aggregateMenu._volume._volumeMenu._input._maybeShowInput = StatusVolume.InputStreamSlider.prototype._maybeShowInput;
        this._updateVisibility();
        super.destroy();
    }

    getDefaultIcon() {
        return 'audio-input-microphone';
    }

    getDefaultStream() {
        return Main.panel.statusArea.aggregateMenu._volume._control.get_default_source();
    }

    getDeviceById(id) {
        return Main.panel.statusArea.aggregateMenu._volume._control.lookup_input_id(id);
    }
});

const PopupVolumeOutputSubMenuMenuItem = GObject.registerClass(
class PopupVolumeOutputSubMenuMenuItem extends PopupVolumeSubMenuMenuItem {
    _init() {
        super._init('output');
    }

    changeDevice(uidevice) {
        Main.panel.statusArea.aggregateMenu._volume._control.change_output(uidevice);
    }

    getDefaultIcon() {
        return 'audio-card';
    }

    getDefaultStream() {
        return Main.panel.statusArea.aggregateMenu._volume._control.get_default_sink();
    }

    getDeviceById(id) {
        return Main.panel.statusArea.aggregateMenu._volume._control.lookup_output_id(id);
    }
});

class VolumeIndicator {
    constructor() {
        this._itemOutput = new PopupVolumeOutputSubMenuMenuItem();
        Main.panel.statusArea.aggregateMenu._volume._volumeMenu.addMenuItem(this._itemOutput, 1);
        this._itemInput = new PopupVolumeInputSubMenuMenuItem();
        Main.panel.statusArea.aggregateMenu._volume._volumeMenu.addMenuItem(this._itemInput, 3);
    }

    destroy() {
        this._itemInput.destroy();
        this._itemOutput.destroy();
    }
}

class SystemIndicator {
    constructor() {
        this._itemTerminal = new PopupTerminalMenuItem();
        Main.panel.statusArea.aggregateMenu._system.menu.addMenuItem(this._itemTerminal, 2);
        this._separator = new PopupMenu.PopupSeparatorMenuItem();
        Main.panel.statusArea.aggregateMenu._system.menu.addMenuItem(this._separator, 2);
        this._itemPreferences = new PopupPreferencesSubMenuMenuItem();
        Main.panel.statusArea.aggregateMenu._system.menu.addMenuItem(this._itemPreferences, 2);
        this._itemSettings = new PopupSettingsMenuItem();
        Main.panel.statusArea.aggregateMenu._system.menu.addMenuItem(this._itemSettings, 2);
        this._itemHibernation = new PopupHibernationMenuSection();
        Main.panel.statusArea.aggregateMenu._system._sessionSubMenu.menu.addMenuItem(this._itemHibernation, 1);
    }

    destroy() {
        this._itemHibernation.destroy();
        this._itemSettings.destroy();
        this._itemPreferences.destroy();
        this._separator.destroy();
        this._itemTerminal.destroy();
    }
}

class SystemMenuButton {
    constructor() {
        Main.panel.statusArea.aggregateMenu.remove_child(Main.panel.statusArea.aggregateMenu._indicators);
        this._hbox = new St.BoxLayout();
        this._hbox.add_child(Main.panel.statusArea.aggregateMenu._indicators);
        this._indicator = new St.BoxLayout({ style_class: 'panel-status-indicators-box' });
        this._indicator.add_child(new St.Icon({ icon_name: 'system-shutdown-symbolic', style_class: 'system-status-icon' }));
        this._indicator.add_child(PopupMenu.arrowIcon(St.Side.BOTTOM));
        this._hbox.add_child(this._indicator);
        Main.panel.statusArea.aggregateMenu.add_child(this._hbox);
        this._settings = ExtensionUtils.getSettings();
        this._settingsChangedId = this._settings.connect('changed::%s'.format(KEY_SHOW_SYSTEMMENU), this._onChangedShowSystemMenu.bind(this));
        this._onChangedShowSystemMenu();
        this._volume = new VolumeIndicator();
        this._system = new SystemIndicator();
    }

    _onChangedShowSystemMenu() {
        let showSystemMenu = this._settings.get_enum(KEY_SHOW_SYSTEMMENU);
        Main.panel.statusArea.aggregateMenu._indicators.visible = showSystemMenu === SYSTEMMENU_DEFAULT;
        this._indicator.visible = showSystemMenu === SYSTEMMENU_ICON;
    }

    destroy() {
        this._system.destroy();
        this._volume.destroy();
        this._settings.disconnect(this._settingsChangedId);
        Main.panel.statusArea.aggregateMenu.remove_child(this._hbox);
        this._hbox.remove_child(Main.panel.statusArea.aggregateMenu._indicators);
        this._hbox.destroy_all_children();
        this._hbox.destroy();
        Main.panel.statusArea.aggregateMenu.add_child(Main.panel.statusArea.aggregateMenu._indicators);
        Main.panel.statusArea.aggregateMenu._indicators.show();
    }
}

function getExtension() {
    return new SystemMenuButton();
}
